/*****************************************************************************
* wanpipe_sppp.c	
* 		WANPIPE PPP/CHDLC Protocol Module. Main code.
*
* Author:	Nenad Corbic
*
* Copyright:	(c)2004 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
*	This module:
*		This module is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
*/
#include "wanpipe_sppp.h"

unsigned int debug = 0;

void *wp_register_ppp_prot(void *link_ptr, 
		 	    char *devname, 
		            void *cfg_ptr, 
			    wplip_prot_reg_t *prot_reg)
{
	wp_sppp_t *sppp;

	FUNC_BEGIN();

	if ((sppp = wpsppp_create_cb()) == NULL){
		wpabs_debug_event( "SPPP: Failed to create cb %s\n",
				devname);
		return NULL;
	}

	
	wpabs_memcpy(&sppp->reg,prot_reg,sizeof(wplip_prot_reg_t));

		
	sppp->link_dev = link_ptr;
	sppp->pp_flags &= ~PP_CISCO;

	FUNC_END();
		
	return sppp;
}

void *wp_register_chdlc_prot(void *link_ptr, 
		 	     char *devname, 
		             void *cfg_ptr, 
			     wplip_prot_reg_t *prot_reg)
{
	wp_sppp_t *sppp;

	FUNC_BEGIN();

	if ((sppp = wpsppp_create_cb()) == NULL){
		wpabs_debug_event( "SPPP: Failed to create cb %s\n",
				devname);
		return NULL;
	}

	
	wpabs_memcpy(&sppp->reg,prot_reg,sizeof(wplip_prot_reg_t));

		
	sppp->link_dev = link_ptr;
	sppp->pp_flags |= (PP_CISCO|PP_KEEPALIVE);

	FUNC_END();
		
	return sppp;
}



int wp_unregister_sppp_prot(void *prot_ptr)
{	
	wp_sppp_t *sppp =(wp_sppp_t*)prot_ptr;

	FUNC_BEGIN();

	sppp->link_dev=NULL;
	
	wpsppp_free_cb(sppp);

	FUNC_END();
	return 0;
}


void *wp_register_sppp_chan(void *if_ptr, 
			    void *prot_ptr, 
		            char *devname, 
			    void *cfg_ptr,
			    unsigned char type)
{
	wp_sppp_t *sppp =(wp_sppp_t*)prot_ptr;
	wan_sppp_if_conf_t *cfg = (wan_sppp_if_conf_t *)cfg_ptr;

	wpabs_strncpy(sppp->name,devname,sizeof(sppp->name)-1);
	wpabs_memcpy(&sppp->cfg,cfg,sizeof(wan_sppp_if_conf_t));

	wp_sppp_attach(sppp);
	sppp->dev = if_ptr;

	sppp->type=type;
	
	return sppp;
}


int wp_unregister_sppp_chan(void *chan_ptr)
{
	wp_sppp_t *sppp =(wp_sppp_t*)chan_ptr;
	wp_sppp_detach(sppp);
	sppp->dev=NULL;
	return 0;
}


int wp_sppp_open(void *sppp_ptr)
{
	wp_sppp_t *sp =(wp_sppp_t*)sppp_ptr;

	wp_sppp_close(sp);
	if (!(sp->pp_flags & PP_CISCO)) {
		wpsppp_lcp_open (sp);
	}
	sp->pp_link_state = SPPP_LINK_DOWN;
	return 0;
}

/**
 *	wp_sppp_close - close down a synchronous PPP or Cisco HDLC link
 *	@dev: The network device to drop the link of
 *
 *	This drops the logical interface to the channel. It is not
 *	done politely as we assume we will also be dropping DTR. Any
 *	timeouts are killed.
 */

int wp_sppp_close(void *sppp_ptr)
{
	wp_sppp_t *sp =(wp_sppp_t*)sppp_ptr;

	sp->pp_link_state = SPPP_LINK_DOWN;
	sp->lcp.state = LCP_STATE_CLOSED;
	sp->ipcp.state = IPCP_STATE_CLOSED;
	wpsppp_clear_timeout (sp);
	wpsppp_auth_clear_timeout(sp);

	wpsppp_cp_send (sp, PPP_IPCP, IPCP_TERM_REQ, 0, 0, 0);

	return 0;
}

int wp_sppp_rx(void *sppp_ptr, void *skb)
{
	FUNC_BEGIN();
	wp_sppp_input(sppp_ptr, skb);
	FUNC_END();
	return 0;
}

int wp_sppp_timer(void *sppp_ptr, unsigned int *timeout, unsigned int carrier_reliable)
{
	wp_sppp_t *sppp = (wp_sppp_t*)sppp_ptr;
	unsigned int period = wpabs_get_systemticks();
	int kick_bh=1;
	
	wpabs_debug_test("%s: TIMER: Timeout=%i Auth=%i KeepAlive=%i\n",
			sppp->name,(sppp->pp_timer >0),
			(sppp->pp_auth_timer >0),(sppp->sppp_keepalive_timer >0));

	DBG_CHDLC("%s: TIMER: Timeout=%u Auth=%u KeepAlive=%u\n",
			sppp->name,
			sppp->pp_timer,
			sppp->pp_auth_timer,
			sppp->sppp_keepalive_timer);

	if (!carrier_reliable){
		DBG_CHDLC("carrier is not reliable\n");
		if (sppp->pp_link_state == SPPP_LINK_UP){
			DBG_CHDLC("sppp->pp_link_state == SPPP_LINK_UP\n");
			wp_sppp_reopen(sppp);
			wpsppp_disconnect_indication(sppp,0);
		}
	}
	
	if (sppp->pp_timer){
		DBG_CHDLC("there is sppp->pp_timer\n");
		if ((period-sppp->pp_timer) >= sppp->cfg.pp_timer){
			DBG_CHDLC("%s: PPP Timeout Timer\n", __FUNCTION__);
			wpabs_debug_test("%s: PPP Timeout Timer\n", __FUNCTION__);
			sppp->pp_timer=0;
			wpsppp_cp_timeout(sppp);
			kick_bh=1;
		}
	}

	if (sppp->pp_auth_timer){
		if ((period-sppp->pp_auth_timer) >= sppp->cfg.pp_auth_timer){
			sppp->pp_auth_timer=0;
			wpabs_debug_event("%s: PPP Auth Timer\n",__FUNCTION__);
			kick_bh=1;
		}
	}

	if (sppp->sppp_keepalive_timer){
		DBG_CHDLC("there is sppp->sppp_keepalive_timer\n");

		DBG_CHDLC("period: %d, sppp->sppp_keepalive_timer: %d, sppp->cfg.sppp_keepalive_timer: %d\n",
			period, sppp->sppp_keepalive_timer, sppp->cfg.sppp_keepalive_timer);

		if ((period-sppp->sppp_keepalive_timer) >= sppp->cfg.sppp_keepalive_timer){
			sppp->sppp_keepalive_timer=0;
			wpabs_debug_test("%s: PPP Keepalive Timer\n",__FUNCTION__);
			wpsppp_keepalive(sppp);
			sppp->sppp_keepalive_timer=wpabs_get_systemticks();
			kick_bh=1;
		}
	} else {
		/* In case of no keepalives */
		if (sppp->pp_flags&PP_CISCO) {
		       	if (carrier_reliable) {
	       		 	if (sppp->pp_link_state==SPPP_LINK_DOWN) {
					sppp->pp_loopcnt = 0; 
					sppp->pp_link_state=SPPP_LINK_UP;
					wpabs_debug_event( "%s: protocol up\n", sppp->name);
					wpsppp_connect_indication(sppp, 0);
				}           
			} else {
                         	if (sppp->pp_link_state==SPPP_LINK_UP) {
					sppp->pp_loopcnt = 0;
					sppp->pp_link_state=SPPP_LINK_DOWN;
					wpabs_debug_event( "%s: protocol down\n", sppp->name);
					wpsppp_disconnect_indication(sppp, 0);
				} 
			}
		}
	}

	if (sppp->dynamic_event){
		sppp->dynamic_event=0;
		if (sppp->link_dev && sppp->reg.tx_link_down){
			sppp->reg.kick_task(sppp->link_dev);
		}
	}

	*timeout=wpabs_get_hz();

	return 0;
}

int wp_sppp_task(void *sppp_ptr)
{
	wp_sppp_t *sppp = (wp_sppp_t*)sppp_ptr;

	wpsppp_bh (sppp);

	return 0;
}

/*
 *	Handle transmit packets.
 */
 
static int wp_sppp_tx_hdr(void *sp_ptr, void *skb, unsigned short type)
{
	struct ppp_header *h;
	struct sppp *sp = (struct sppp *)sp_ptr;

	if (wpabs_skb_headroom(skb) < sizeof(struct ppp_header)){
		/* Allowed to fail here, recovery is done 
		   outside this function */
		return -1;
	}
	
	wpabs_skb_push(skb,sizeof(struct ppp_header));
	h=(struct ppp_header *)wpabs_skb_data(skb);
	
	if(sp->pp_flags&PP_CISCO)
	{
		h->address = CISCO_UNICAST;
		h->control = 0;
		if (type == WPLIP_IPV6) {
			h->protocol = wpabs_htons(CISCO_IPV6);
		} else {
			h->protocol = wpabs_htons(CISCO_IP);
		}
	}
	else
	{
		h->address = PPP_ALLSTATIONS;
		h->control = PPP_UI;
		if (type == WPLIP_IPV6) {
			h->protocol = wpabs_htons(PPP_IPV6);
		} else {
			h->protocol = wpabs_htons(PPP_IP);
		}
		sp->packet_stats.tx_ip++;
	}

	
	return sizeof(struct ppp_header);
}

int wp_sppp_tx(void *sp_ptr, void *skb, int type)
{
	struct sppp *sp = (struct sppp *)sp_ptr;
	int len,err;

	len=wp_sppp_tx_hdr(sp_ptr,skb,type);
	if (len < 0){
		unsigned char *buf;
		void *nskb = wpabs_skb_alloc(wpabs_skb_len(skb)+64);
		if (!nskb){
			wpabs_debug_event("SPPP/CHDLC TX HDR FAILED len=%i type=%i: No Mem\n",len,type);
			return -1;
		}
		
		len=wp_sppp_tx_hdr(sp_ptr,nskb,type);
		if (len<0){
			wpabs_skb_free(nskb);
			wpabs_debug_event("SPPP/CHDLC TX HDR FAILED len=%i type=%i\n",len,type);
			return -1;
		}
		
		buf=wpabs_skb_put(nskb,wpabs_skb_len(skb));
		wpabs_memcpy(buf,wpabs_skb_data(skb),wpabs_skb_len(skb));
		
		if (sp->dev && sp->reg.tx_chan_down){
			err=sp->reg.tx_chan_down(sp->dev,nskb);
			if (err){
				/* The nskb failed to tx, we must return
				* the original skb back to the stack, thus
				* free the nskb */
				wpabs_skb_free(nskb);
				return err;
			}    
			
			/* The nskb packet was transmitted ok, we must
			* free the original skb as if original skb was 
			* transmitted */ 
			wpabs_skb_free(skb);
			return err;
		}else{
			/* Absolutely invalid, Critical Error */
			wpabs_debug_event("%s: SPPP: Critical error link dev=%p tx_chan_down=%p\n",
				__FUNCTION__,sp->link_dev,sp->reg.tx_chan_down);
			wpabs_skb_free(nskb);  
			return -1;
		}
	}              	
	
	if (sp->dev && sp->reg.tx_chan_down){
		err=sp->reg.tx_chan_down(sp->dev,skb);
		if (err){
			wpabs_skb_pull(skb,len);
		}
		return err;
	}else{
		wpabs_debug_event("%s: SPPP: Critical error link dev=%p tx_chan_down=%p\n",
			__FUNCTION__,sp->link_dev,sp->reg.tx_chan_down);
		
	}
	return -1;
}


int wp_sppp_pipemon(void *sp_ptr, int cmd, int addr, 
		    unsigned char* data, unsigned int *len)
{	
	struct sppp *sp = (struct sppp *)sp_ptr;

	*len=1;

	if (!(sp->pp_flags & PP_CISCO)) {

		switch (cmd){

		case PPP_READ_CONFIG:
		case PPP_READ_CODE_VERSION:
			break;

		case PPP_READ_AUTH:
			wpabs_memcpy(data,&sp->myauth,sizeof(sp->myauth));
			*len=sizeof(sp->myauth);
			break;

		case PPP_FLUSH_STATISTICS:	
		case PPP_READ_STATISTICS:
			break;

		case PPP_FLUSH_PACKET_STATS:
			wpabs_memset(&sp->packet_stats,0,sizeof(ppp_pkt_stats_t));
			break;

		case PPP_READ_PACKET_STATS:
			/* total of rx LCP */
			sp->packet_stats.rx_lcp = sp->lcp_stats.rx_unknown +
				sp->lcp_stats.rx_conf_rqst + sp->lcp_stats.rx_conf_ack +
				sp->lcp_stats.rx_conf_nak + sp->lcp_stats.rx_conf_rej +
				sp->lcp_stats.rx_term_rqst + sp->lcp_stats.rx_term_ack +
				sp->lcp_stats.rx_code_rej + sp->lcp_stats.rx_proto_rej +
				sp->lcp_stats.rx_echo_rqst + sp->lcp_stats.rx_echo_reply +
				sp->lcp_stats.rx_disc_rqst;

			/* total of tx LCP */
			sp->packet_stats.tx_lcp = sp->lcp_stats.tx_conf_rqst +
				sp->lcp_stats.tx_conf_ack + sp->lcp_stats.tx_conf_nak +
				sp->lcp_stats.tx_conf_rej + sp->lcp_stats.tx_term_rqst +
				sp->lcp_stats.tx_term_ack + sp->lcp_stats.tx_code_rej +
				sp->lcp_stats.tx_proto_rej + sp->lcp_stats.tx_echo_rqst +
				sp->lcp_stats.tx_echo_reply + sp->lcp_stats.tx_disc_rqst;

			/* total of rx IPCP */
			sp->packet_stats.rx_ipcp = 
				sp->prot_stats.rx_unknown + sp->prot_stats.rx_rej_badid +
				sp->prot_stats.rx_conf_ack + sp->prot_stats.rx_conf_nak +
				sp->prot_stats.rx_conf_rej + sp->prot_stats.rx_term_rqst +
				sp->prot_stats.rx_term_ack + sp->prot_stats.rx_code_rej;

			/* total of tx IPCP */
			sp->packet_stats.tx_ipcp = sp->prot_stats.tx_conf_rqst +
				sp->prot_stats.tx_conf_ack + sp->prot_stats.tx_conf_nak +
				sp->prot_stats.tx_conf_rej + sp->prot_stats.tx_term_rqst +
				sp->prot_stats.tx_term_ack + sp->prot_stats.tx_code_rej;

			/* total of rx CHAP */
			sp->packet_stats.rx_chap = sp->chap_stats.rx_unknown +
				sp->chap_stats.rx_challenge + sp->chap_stats.rx_response +
				sp->chap_stats.rx_success + sp->chap_stats.rx_failure +	
				sp->chap_stats.rx_too_large + sp->chap_stats.rx_bad_peerid +
				sp->chap_stats.rx_bad_passwd + sp->chap_stats.rx_bad_md5 +
				sp->chap_stats.rx_bad_resp;

			/* total of tx CHAP */
			sp->packet_stats.tx_chap = sp->chap_stats.tx_challenge +
				sp->chap_stats.tx_response + sp->chap_stats.tx_success +
				sp->chap_stats.tx_failure;

			/* total of rx PAP */
			sp->packet_stats.rx_pap = sp->pap_stats.rx_unknown +
				sp->pap_stats.rx_unknown + sp->pap_stats.rx_auth_rqst +
				sp->pap_stats.rx_auth_ack + sp->pap_stats.rx_auth_nak +
				sp->pap_stats.rx_too_large + sp->pap_stats.rx_bad_peerid +
				sp->pap_stats.rx_bad_passwd;

			/* total of tx PAP */
			sp->packet_stats.tx_pap = sp->pap_stats.tx_auth_rqst +
				sp->pap_stats.tx_auth_ack + sp->pap_stats.tx_auth_nak;

			wpabs_memcpy(data,&sp->packet_stats,sizeof(ppp_pkt_stats_t));
			*len=sizeof(ppp_pkt_stats_t);
			break;
			
		case PPP_FLUSH_LCP_STATS:
			wpabs_memset(&sp->lcp_stats,0,sizeof(ppp_lcp_stats_t));
			break;

		case PPP_READ_LCP_STATS:
			wpabs_memcpy(data,&sp->lcp_stats,sizeof(ppp_lcp_stats_t));
			*len=sizeof(ppp_lcp_stats_t);
			break;

		case PPP_FLUSH_LPBK_STATS:
			wpabs_memset(&sp->lpbk_stats,0,sizeof(ppp_lpbk_stats_t));
			break;
		
		case PPP_READ_LPBK_STATS:
			wpabs_memcpy(data,&sp->lpbk_stats,sizeof(ppp_lpbk_stats_t));
			*len=sizeof(ppp_lpbk_stats_t);
			break;

		case PPP_FLUSH_IPCP_STATS:
		case PPP_FLUSH_IPXCP_STATS:
			wpabs_memset(&sp->prot_stats,0,sizeof(ppp_prot_stats_t));
			break;

		case PPP_READ_IPCP_STATS:
		case PPP_READ_IPXCP_STATS:
			wpabs_memcpy(data,&sp->prot_stats,sizeof(ppp_prot_stats_t));
			*len=sizeof(ppp_prot_stats_t);
			break;
			
		case PPP_FLUSH_PAP_STATS:
			wpabs_memset(&sp->pap_stats,0,sizeof(ppp_pap_stats_t));
			break;

		case PPP_READ_PAP_STATS:
			wpabs_memcpy(data,&sp->pap_stats,sizeof(ppp_pap_stats_t));
			*len=sizeof(ppp_pap_stats_t);
			break;

		case PPP_FLUSH_CHAP_STATS:
			wpabs_memset(&sp->chap_stats,0,sizeof(ppp_chap_stats_t));
			break;

		case PPP_READ_CHAP_STATS:
			wpabs_memcpy(data,&sp->chap_stats,sizeof(ppp_chap_stats_t));
			*len=sizeof(ppp_chap_stats_t);
			break;
		
		case PPP_GET_CONNECTION_INFO:
			break;

		case PPP_GET_LINK_STATUS:
			/* protocol Connected/Disconnected (link Operative/Inoperative) */
			data[0] = sp->pp_link_state;
			*len=1;
			break;
			
		default: 
			return -1;
		}

	}else{
		DBG_CHDLC("%s: CHDLC CMD 0x%X\n",sp->name,cmd);  
		
		switch (cmd)
		{
#if 0
		case READ_GLOBAL_CONFIGURATION:
		case READ_CHDLC_CODE_VERSION:
			break;
#endif
		case READ_CHDLC_CONFIGURATION:
			wpabs_memcpy(data,&sp->original_cfg,sizeof(wan_sppp_if_conf_t));
			*len=sizeof(wan_sppp_if_conf_t);
			break;

		case FLUSH_CHDLC_OPERATIONAL_STATS:
			wpabs_memset(&sp->chdlc_stats,0,sizeof(CHDLC_OPERATIONAL_STATS_STRUCT));
			break;

		case READ_CHDLC_OPERATIONAL_STATS:
			wpabs_memcpy(data,&sp->chdlc_stats,sizeof(CHDLC_OPERATIONAL_STATS_STRUCT));
			*len=sizeof(CHDLC_OPERATIONAL_STATS_STRUCT);
			break;
		
		default: 
			return -1;
		}

	}

	return 0;
}


/*======================================================
 * wpsppp_data_transmit
 *
 * Transmitt the ppp packet to the lower layer,
 * vi its master (tx_dev) device.
 * 
 */
int wpsppp_data_transmit(wp_sppp_t *sppp, void *skb)
{
	wpabs_debug_test("%s: %s() Len=%i\n",
			sppp->name,__FUNCTION__,wpabs_skb_len(skb));

	if (sppp->link_dev && sppp->reg.tx_link_down){
		return sppp->reg.tx_link_down(sppp->link_dev,skb);
	}else{
		wpabs_debug_event("%s: Critical error link dev=%p tx_link_down=%p\n",
				__FUNCTION__,sppp->link_dev,sppp->reg.tx_link_down);
				
	}
	
	return 1;
}

/*======================================================
 * wpsppp_data_indication
 *
 * Transmitt the lapb packet to the lower layer,
 * vi its master (tx_dev) device.
 * 
 */
int wpsppp_data_indication(wp_sppp_t *sppp, void *skb, int type)
{
	wpabs_debug_test("%s: %s() Len=%i\n",
			sppp->name,__FUNCTION__,wpabs_skb_len(skb));

	if (sppp->dev){
		return sppp->reg.rx_up(sppp->dev,skb,type);
	}

	wpabs_skb_free(skb);

	return 1;
}



/**********************************************************
 * Private Functions
 **********************************************************/


/*======================================================
 * sppp_connect_indication
 *
 * Indicate to the upper layer (x25) that the
 * lapb is layer is connected and that it should
 * start its connection process.
 */
void wpsppp_connect_indication(wp_sppp_t *sppp, int reason)
{
	wpabs_debug_test("%s: %s() Reason=%i\n",
			sppp->name,__FUNCTION__,reason);

	if (sppp->dev && sppp->reg.chan_set_state){
		sppp->reg.chan_set_state(sppp->dev,WAN_CONNECTED,NULL, 0);
	}
}


/*======================================================
 * sppp_disconnect_indication
 *
 * Indicate to the upper layer (x25) that the
 * lapb is layer is disconnected and that it should
 * drop the connection.
 */
void wpsppp_disconnect_indication(wp_sppp_t *sppp, int reason)
{
	wpabs_debug_test("%s: %s() Reason=%i\n",
			sppp->name,__FUNCTION__,reason);

	if (sppp->dev && sppp->reg.chan_set_state){
		sppp->reg.chan_set_state(sppp->dev,WAN_DISCONNECTED,NULL,0);
	}
}


unsigned int wpsppp_get_ipv4addr(wp_sppp_t *sppp, int type)
{
	if (sppp->dev && sppp->reg.get_ipv4_addr){
		return sppp->reg.get_ipv4_addr(sppp->dev,type);
	}
	return 0;
}

int wpsppp_set_ipv4addr(wp_sppp_t *sppp,
		 	unsigned int local,
			unsigned int remote,
			unsigned int netmask,
			unsigned int dns)
{
	if (sppp->dev && sppp->reg.set_ipv4_addr){
		return sppp->reg.set_ipv4_addr(sppp->dev,
					      local,
				              remote,
					      netmask,
					      dns);
	}
	return 0;
}

int wpsppp_set_gateway(wp_sppp_t *sppp)
{
	return 0;
}

